import bz2
import datetime
import json
import lzma
import os

from eveuniverse.models import EveConstellation, EveSolarSystem

from django.db import IntegrityError

from allianceauth.eveonline.models import EveFactionInfo
from allianceauth.services.hooks import get_extension_logger

from incursions.models import Incursion, IncursionInfluence

from .static_data import incursion_constellations

logger = get_extension_logger(__name__)


EVEREF_STAGING_FOLDER = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'data.everef.net/incursions/history/')


def import_staging_history(dir: str = EVEREF_STAGING_FOLDER):
    for entry in os.scandir(dir):
        if entry.name.startswith('incursions-') and entry.is_file():
            with bz2.open(entry) as f:
                for incursion in f.read().decode():
                    print(incursion)
        elif entry.is_dir():
            import_staging_history(entry)
# [{"constellation_id":20000651,"faction_id":500019,"has_boss":false,"infested_solar_systems":[30004459,30004460,30004461,30004462,30004463,30004464],"influence":0.0,"staging_solar_system_id":30004464,"state":"mobilizing","type":"Incursion"},{"constellation_id":20000520,"faction_id":500019,"has_boss":true,"infested_solar_systems":[30003565,30003566,30003567,30003568,30003569,30003570,30003571,30003572],"influence":1.0,"staging_solar_system_id":30003572,"state":"mobilizing","type":"Incursion"},{"constellation_id":20000173,"faction_id":500019,"has_boss":false,"infested_solar_systems":[30001184,30001185,30001186,30001187,30001182,30001183],"influence":0.0,"staging_solar_system_id":30001186,"state":"mobilizing","type":"Incursion"},{"constellation_id":20000618,"faction_id":500019,"has_boss":false,"infested_solar_systems":[30004224,30004225,30004226,30004227,30004228,30004229,30004222,30004223],"influence":0.32516667656600473,"staging_solar_system_id":30004227,"state":"established","type":"Incursion"}]


def import_staging_backfill(dir: str = EVEREF_STAGING_FOLDER):
    with lzma.open(os.path.join(EVEREF_STAGING_FOLDER, "backfills/eve-incursions-de-2023-10-12.json.xz")) as f:
        data = f.read()
        for incursion in json.loads(data):
            eve_constellation = EveConstellation.objects.get_or_create_esi(id=incursion['spawn']["constellation"]["id"])[0]
            if incursion['spawn']['endedAt'] is None:
                # dont handle any ongoing incursions.
                continue
            try:
                i = Incursion.objects.get(
                    established_timestamp__date=datetime.datetime.strptime(str(incursion['spawn']["establishedAt"]), "%Y-%m-%dT%H:%M:%S.%fZ").date(),
                    ended_timestamp__date=datetime.datetime.strptime(str(incursion['spawn']["endedAt"]), "%Y-%m-%dT%H:%M:%S.%fZ").date(),
                    constellation=eve_constellation)
            except Incursion.DoesNotExist:
                try:
                    i = Incursion.objects.create(
                        constellation=eve_constellation,
                        faction=EveFactionInfo.objects.get_or_create(faction_id=500019)[0],
                        staging_solar_system=EveSolarSystem.objects.get(name=incursion_constellations[eve_constellation.name]["Staging"]),
                        state=Incursion.States.ENDED,
                        established_timestamp=datetime.datetime.strptime(str(incursion['spawn']["establishedAt"]), "%Y-%m-%dT%H:%M:%S.%fZ"),
                        ended_timestamp=datetime.datetime.strptime(str(incursion['spawn']["endedAt"]), "%Y-%m-%dT%H:%M:%S.%fZ")
                    )
                except EveSolarSystem.DoesNotExist:
                    logger.error(eve_constellation.name)
            # Save an extra timestamps we can gather from this data
            try:
                if incursion['state'] == "Mobilizing":
                    i.mobilizing_timestamp = datetime.datetime.strptime(str(incursion['date']), "%Y-%m-%dT%H:%M:%S.%fZ")
                    i.save()
                elif incursion['state'] == "Withdrawing":
                    i.withdrawing_timestamp = datetime.datetime.strptime(str(incursion['date']), "%Y-%m-%dT%H:%M:%S.%fZ")
                    i.save()
            except Exception as e:
                logger.exception(e)
            # Moving onto IncursionInfluence before we close the loop
            for ilog in incursion["spawn"]["influenceLogs"]:
                try:
                    IncursionInfluence.objects.create(
                        incursion=i,
                        influence=ilog['influence'],
                        timestamp=datetime.datetime.strptime(str(ilog["date"]), "%Y-%m-%dT%H:%M:%S.%fZ"))
                except IntegrityError:
                    # We already have this data
                    pass

#   {
#     "id": "1",
#     "date": "2015-02-06T18:13:45.000Z",
#     "state": "Established",
#     "spawn": {
#       "id": "12",
#       "establishedAt": "2015-02-06T18:13:45.000Z",
#       "endedAt": "2015-02-09T16:00:01.000Z",
#       "state": "Ended",
#       "constellation": {
#         "id": "20000011"
#       },
#       "influenceLogs": [
#         {
#           "id": "293702",
#           "date": "2021-10-30T09:00:00.000Z",
#           "influence": 0
#         },
#         {
#           "id": "293707",
#           "date": "2021-10-30T10:00:00.000Z",
#           "influence": 0
#         },
#         {
#           "id": "293712",
#           "date": "2021-10-30T11:00:00.000Z",
#           "influence": 0.048
#         },
#         {
#           "id": "293717",
#           "date": "2021-10-30T12:00:00.000Z",
#           "influence": 0.175333
#         },
#     }
#   },



from django.contrib import admin

from allianceauth.services.hooks import get_extension_logger

from .models import (
    Focus, Incursion, IncursionInfluence, IncursionsConfig, Webhook,
)

logger = get_extension_logger(__name__)


@admin.register(IncursionsConfig)
class IncursionsConfigAdmin(admin.ModelAdmin):
    filter_horizontal = ["status_webhooks", ]


@admin.register(Incursion)
class IncursionAdmin(admin.ModelAdmin):
    list_display = ["constellation", "state", "established_timestamp",
                    "mobilizing_timestamp", "withdrawing_timestamp", "ended_timestamp"]
    list_filter = ["state", "has_boss"]
    filter_horizontal = ["infested_solar_systems", ]


@admin.register(Webhook)
class WebhookAdmin(admin.ModelAdmin):
    list_display = ("name", "url")


@admin.register(IncursionInfluence)
class IncursionInfluenceAdmin(admin.ModelAdmin):
    list_display = ("incursion", "timestamp", "influence")


@admin.register(Focus)
class FocusAdmin(admin.ModelAdmin):
    list_display = ["incursion",]

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "incursion":
            kwargs["queryset"] = Incursion.objects.exclude(state=Incursion.States.ENDED)
        return super().formfield_for_foreignkey(db_field, request, **kwargs)



from datetime import timedelta
from typing import Union

from discord import SyncWebhook
from eveuniverse.models import EveConstellation, EveSolarSystem
from solo.models import SingletonModel

from django.db import models
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import EveFactionInfo


class Incursion(models.Model):

    class States(models.TextChoices):
        """"
        Incursion Phases
        """
        ESTABLISHED = 'established', _("Incursion established")  # Up to 5 Days
        MOBILIZING = 'mobilizing', _("Incursion mobilized")  # 48 Hours
        WITHDRAWING = 'withdrawing', _("Incursion withdrawing")  # 24 Hours
        # Our custom phase to archive finished incursions
        ENDED = 'ended', _("Ended")

    class Types(models.TextChoices):
        """
        I Think this used to be used for Trig Incursions etc, leaving the option in place to expand it later
        """
        INCURSION = 'Incursion', _("Sansha Incursion")

    constellation = models.ForeignKey(EveConstellation, verbose_name=_(
        "The constellation id in which this incursion takes place"), on_delete=models.CASCADE)
    faction = models.ForeignKey(EveFactionInfo, verbose_name=_(
        "The attacking Faction"), on_delete=models.CASCADE)
    has_boss = models.BooleanField(_("Whether the final encounter has boss or not"), default=False)
    infested_solar_systems = models.ManyToManyField(EveSolarSystem, verbose_name=_(
        "A list of infested solar system ids that are a part of this incursion"), related_name="+")
    staging_solar_system = models.ForeignKey(EveSolarSystem, verbose_name=_(
        "Staging solar system for this incursion"), on_delete=models.CASCADE, related_name="+")
    state = models.CharField(
        _("The state of this incursion"),
        max_length=50, choices=States.choices, default=States.ESTABLISHED)
    type = models.CharField(
        _("The type of this incursion"), max_length=50,
        choices=Types.choices, default=Types.INCURSION)

    established_timestamp = models.DateTimeField(
        _("Established"), auto_now=False, auto_now_add=False, blank=True, null=True)
    mobilizing_timestamp = models.DateTimeField(
        _("Mobilizing"), auto_now=False, auto_now_add=False, blank=True, null=True)
    withdrawing_timestamp = models.DateTimeField(
        _("Withdrawing"), auto_now=False, auto_now_add=False, blank=True, null=True)
    ended_timestamp = models.DateTimeField(
        _("Ended"), auto_now=False, auto_now_add=False, blank=True, null=True)

    class Meta:
        verbose_name = _("Incursion")
        verbose_name_plural = _("Incursions")

    def __str__(self) -> str:
        return f"{self.constellation.name}"

    @property
    def established_utlization(self) -> Union[float, bool]:
        if self.established_timestamp and self.mobilizing_timestamp:
            return (self.established_timestamp - self.mobilizing_timestamp) / timedelta(days=5)
        else:
            return False

    @property
    def mobilizing_utlization(self) -> Union[float, bool]:
        if self.mobilizing_timestamp and self.withdrawing_timestamp:
            return (self.mobilizing_timestamp - self.withdrawing_timestamp) / timedelta(days=2)
        else:
            return False

    @property
    def security_string(self) -> str:
        if self.staging_solar_system.is_high_sec:
            return "High-Sec"
        elif self.staging_solar_system.is_low_sec:
            return "Low-Sec"
        elif self.staging_solar_system.is_null_sec:
            return "Null-Sec"
        else:
            return "Unknown"

    @property
    def is_island(self) -> bool:
        try:
            for system in self.staging_solar_system.route_to(destination=EveSolarSystem.objects.get(30000142)):
                if system.is_high_sec is False:
                    return True
            return False
        except Exception:
            # wtf
            return True  # WHY NOT

    @property
    def influence(self) -> float:
        try:
            IncursionInfluence.objects.filter(incursion=self).latest("timestamp").influence
        except Exception:
            return float(0)


class IncursionInfluence(models.Model):
    incursion = models.ForeignKey(Incursion, verbose_name=_("Incursion"), on_delete=models.CASCADE)
    timestamp = models.DateTimeField(_("Timestamp"), auto_now=False, auto_now_add=False)
    influence = models.FloatField(
        _("Influence of this incursion as a float from 0 to 1"), default=float(0))

    class Meta:
        verbose_name = _("IncursionInfluence")
        verbose_name_plural = _("IncursionInfluences")
        constraints = [
            models.UniqueConstraint(fields=['incursion', 'timestamp'], name="UniqueIncursionInfluenceLogTimestamp"),
        ]

    def __str__(self) -> str:
        return f"{self.incursion} @ {self.timestamp}"


class Webhook(models.Model):
    """Destinations for Relays"""
    url = models.CharField(max_length=200)
    name = models.CharField(max_length=50)

    def __str__(self) -> str:
        return f'"{self.name}"'

    class Meta:
        verbose_name = _('Destination Webhook')
        verbose_name_plural = _('Destination Webhooks')

    def send_embed(self, embed):
        webhook = SyncWebhook.from_url(self.url)
        webhook.send(embed=embed, username="AA Incursions")


class IncursionsConfig(SingletonModel):
    status_webhooks = models.ManyToManyField(
        Webhook, verbose_name=_("Destination Webhook for Incursion Updates"))

    security_high = models.BooleanField(_("Notify on High Security Incursions"))
    security_low = models.BooleanField(_("Notify on Low Security Incursions"))
    security_null = models.BooleanField(_("Notify on Null Security Incursions"))

    def __str__(self) -> str:
        return "AA Incursions Settings"

    class Meta:
        """
        Meta definitions
        """
        default_permissions = ()
        verbose_name = _("AA Incursions Settings")
        verbose_name_plural = _("AA Incursions Settings")


class Focus(SingletonModel):
    incursion = models.ForeignKey(
        Incursion,
        verbose_name=_("Current Focus"),
        on_delete=models.CASCADE, blank=True, null=True)

    def __str__(self) -> str:
        try:
            return f"Current Focus: {self.incursion}"
        except Exception:
            return "No Focus Set"

    class Meta:
        """
        Meta definitions
        """
        default_permissions = ()
        verbose_name = _("Current Focus")
        verbose_name_plural = _("Current Focus")



import logging

from django.db.models.signals import post_save
from django.dispatch import receiver

from incursions.app_settings import INCURSIONS_AUTO_HIGHSEC_STATIC
from incursions.models import Focus, Incursion
from incursions.tasks import (
    incursion_boss_spawned, incursion_ended, incursion_established,
    incursion_mobilizing, incursion_withdrawing,
)

logger = logging.getLogger(__name__)


@receiver(post_save, sender=Incursion)
def incursion(sender, instance: Incursion, created: bool, *args, **kwargs):
    update_fields = kwargs.pop('update_fields', []) or []

    if created is True:
        incursion_established.apply_async(args=[instance.pk])
        if INCURSIONS_AUTO_HIGHSEC_STATIC and instance.staging_solar_system.is_high_sec:
            Focus.get_solo().incursion = instance

    if 'state' in update_fields:
        if instance.state == Incursion.States.ESTABLISHED:
            # This should have been handled above?
            pass
        elif instance.state == Incursion.States.MOBILIZING:
            incursion_mobilizing.apply_async(args=[instance.pk])
        elif instance.state == Incursion.States.WITHDRAWING:
            incursion_withdrawing.apply_async(args=[instance.pk])
        elif instance.state == Incursion.States.ENDED:
            incursion_ended.apply_async(args=[instance.pk])

    if 'has_boss' in update_fields and instance.has_boss is True:
        incursion_boss_spawned.apply_async(args=[instance.pk])



# Generated by Django 4.0 on 2023-10-31 14:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveuniverse', '0010_alter_eveindustryactivityduration_eve_type_and_more'),
        ('incursions', '0005_incursioninfluence_uniqueincursioninfluencelogtimestamp'),
    ]

    operations = [
        migrations.AlterField(
            model_name='incursion',
            name='infested_solar_systems',
            field=models.ManyToManyField(related_name='+', to='eveuniverse.EveSolarSystem', verbose_name='A list of infested solar system ids that are a part of this incursion'),
        ),
        migrations.AlterField(
            model_name='incursionsconfig',
            name='status_webhooks',
            field=models.ManyToManyField(to='incursions.Webhook', verbose_name='Destination Webhook for Incursion Updates'),
        ),
    ]



# Generated by Django 4.0.10 on 2023-08-04 04:32

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incursions', '0003_focus'),
    ]

    operations = [
        migrations.AlterField(
            model_name='focus',
            name='incursion',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='incursions.incursion', verbose_name='Current Focus'),
        ),
    ]



# Generated by Django 4.0.10 on 2023-06-22 04:52

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incursions', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='incursion',
            name='influence',
        ),
        migrations.AlterField(
            model_name='incursion',
            name='ended_timestamp',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Ended'),
        ),
        migrations.AlterField(
            model_name='incursion',
            name='established_timestamp',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Established'),
        ),
        migrations.AlterField(
            model_name='incursion',
            name='mobilizing_timestamp',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Mobilizing'),
        ),
        migrations.AlterField(
            model_name='incursion',
            name='withdrawing_timestamp',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Withdrawing'),
        ),
        migrations.CreateModel(
            name='IncursionInfluence',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('timestamp', models.DateTimeField(verbose_name='Timestamp')),
                ('influence', models.FloatField(default=0.0, verbose_name='Influence of this incursion as a float from 0 to 1')),
                ('incursion', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='incursions.incursion', verbose_name='Incursion')),
            ],
            options={
                'verbose_name': 'IncursionInfluence',
                'verbose_name_plural': 'IncursionInfluences',
            },
        ),
    ]



# Generated by Django 4.0.10 on 2023-08-04 04:28

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incursions', '0002_remove_incursion_influence_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='Focus',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('incursion', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='incursions.incursion', verbose_name='Current Focus')),
            ],
            options={
                'verbose_name': 'Focus',
                'verbose_name_plural': 'Focus',
                'default_permissions': (),
            },
        ),
    ]



import logging

from discord import AutocompleteContext, Option
from discord.commands import SlashCommandGroup
from discord.embeds import Embed
from discord.ext import commands
from eveuniverse.models import EveConstellation, EveRegion, EveSolarSystem

from django.conf import settings

from incursions import __version__
from incursions.models import Focus, Incursion

logger = logging.getLogger(__name__)


class Incursions(commands.Cog):
    """
    Drifter Wormhole Mapping and Management
    From AA-incursions
    """

    def __init__(self, bot):
        self.bot = bot

    incursion_commands = SlashCommandGroup(
        "incursions", "Incursions", guild_ids=[int(settings.DISCORD_GUILD_ID)])

    async def search_solar_systems(self, ctx: AutocompleteContext):
        return list(EveSolarSystem.objects.filter(name__icontains=ctx.value).values_list('name', flat=True)[:10])

    async def search_constellations(self, ctx: AutocompleteContext):
        return list(EveConstellation.objects.filter(name__icontains=ctx.value).values_list('name', flat=True)[:10])

    async def search_regions(self, ctx: AutocompleteContext):
        return list(EveRegion.objects.filter(name__icontains=ctx.value).values_list('name', flat=True)[:10])

    async def search_incursion_active(self, ctx: AutocompleteContext):
        return list(Incursion.objects.exclude(state=Incursion.States.ENDED).filter(constellation__name__icontains=ctx.value).values_list('constellation__name', flat=True)[:10])

    @incursion_commands.command(name="about", description="About the Incursion Bot", guild_ids=[int(settings.DISCORD_GUILD_ID)])
    async def about(self, ctx):
        """
        All about the bot
        """
        embed = Embed(title="AA Incursions")
        embed.description = "https://gitlab.com/tactical-supremacy/aa-incursions"
        embed.url = "https://gitlab.com/tactical-supremacy/aa-incursions"
        embed.set_thumbnail(url="https://images.evetech.net/types/2192/render?size=128")
        embed.set_footer(
            text="Developed by Ariel Rin")
        embed.add_field(
            name="Version", value=f"{__version__}", inline=False
        )

        return await ctx.respond(embed=embed)

    @incursion_commands.command(name="focus", description="Get information on the current focus", guild_ids=[int(settings.DISCORD_GUILD_ID)])
    async def focus(
        self, ctx,
    ):
        incursion_obj = Focus.get_solo().incursion
        incursion_detail_string = f"""
            {incursion_obj.constellation.name} ({incursion_obj.state})
            Established: {incursion_obj.established_timestamp}
            Mobilized: {incursion_obj.mobilizing_timestamp}
            Withdrawing:{incursion_obj.withdrawing_timestamp}
            Boss Spotted: {incursion_obj.has_boss}
            Influence: {incursion_obj.influence}
            """
        return await ctx.respond(incursion_detail_string)

    @incursion_commands.command(name="set_focus", description="Set the current Focus", guild_ids=[int(settings.DISCORD_GUILD_ID)])
    async def focus_set(
        self, ctx,
        incursion=Option(str, "Constellation", autocomplete=search_incursion_active),
    ):
        focus = Focus.get_solo()
        focus.incursion_id = Incursion.objects.exclude(
            state=Incursion.States.ENDED).get(constellation__name__icontains=incursion)
        focus.save()
        return await ctx.respond("Focus Set")

    @incursion_commands.command(name="incursions", description="List active incursions", guild_ids=[int(settings.DISCORD_GUILD_ID)])
    async def incursions(
        self, ctx,
    ):
        incursion_string = ""
        for incursion in Incursion.objects.exclude(state=Incursion.States.ENDED):
            incursion_string += f"\n{incursion.constellation.name}"
        return await ctx.respond(incursion_string)

    @incursion_commands.command(name="incursion_detail", description="Status of a specific incursion", guild_ids=[int(settings.DISCORD_GUILD_ID)])
    async def incursion_detail(
        self, ctx,
        incursion=Option(str, "Constellation", autocomplete=search_incursion_active),
    ):
        incursion_obj = Incursion.objects.exclude(state=Incursion.States.ENDED).get(
            constellation__name__icontains=incursion)
        incursion_detail_string = f"""
            {incursion_obj.constellation.name} ({incursion_obj.state})
            Established: {incursion_obj.established_timestamp}
            Mobilized: {incursion_obj.mobilizing_timestamp}
            Withdrawing:{incursion_obj.withdrawing_timestamp}
            Boss Spotted: {incursion_obj.has_boss}
            Influence: {incursion_obj.influence}
            """
        return await ctx.respond(incursion_detail_string)


def setup(bot):
    bot.add_cog(Incursions(bot))



# Generated by Django 4.2.10 on 2024-05-11 11:56

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('incursions', '0006_alter_incursion_infested_solar_systems_and_more'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='focus',
            options={'default_permissions': (), 'verbose_name': 'Current Focus', 'verbose_name_plural': 'Current Focus'},
        ),
    ]



# Generated by Django 4.0.10 on 2023-10-13 09:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incursions', '0004_alter_focus_incursion'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='incursioninfluence',
            constraint=models.UniqueConstraint(fields=('incursion', 'timestamp'), name='UniqueIncursionInfluenceLogTimestamp'),
        ),
    ]
